import logging

from django.test import TestCase

from eth_account import Account

from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import get_safe_contract
from gnosis.eth.utils import get_eth_address_with_key

from ..safe_creation_tx import SafeCreationTx
from .safe_test_case import SafeTestCaseMixin
from .utils import generate_valid_s

logger = logging.getLogger(__name__)

LOG_TITLE_WIDTH = 100


class TestSafeCreationTx(SafeTestCaseMixin, TestCase):
    def test_safe_creation_tx_builder(self):
        logger.info("Test Safe Proxy creation without payment".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3

        s = generate_valid_s()

        funder_account = self.ethereum_test_account
        owners = [get_eth_address_with_key()[0] for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreationTx(w3=w3,
                                          owners=owners,
                                          threshold=threshold,
                                          signature_s=s,
                                          master_copy=self.safe_contract_V0_0_1_address,
                                          gas_price=gas_price,
                                          funder=NULL_ADDRESS)

        logger.info("Send %d gwei to deployer %s",
                    w3.fromWei(safe_creation_tx.payment_ether, 'gwei'),
                    safe_creation_tx.deployer_address)

        self.send_tx({
            'to': safe_creation_tx.deployer_address,
            'value': safe_creation_tx.payment_ether
        }, funder_account)

        logger.info("Create proxy contract with address %s", safe_creation_tx.safe_address)

        tx_hash = w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        self.assertEqual(tx_receipt.contractAddress, safe_creation_tx.safe_address)

        deployed_safe_proxy_contract = get_safe_contract(w3, tx_receipt.contractAddress)

        logger.info("Deployer account has still %d gwei left (will be lost)",
                    w3.fromWei(w3.eth.get_balance(safe_creation_tx.deployer_address), 'gwei'))

        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)

    def test_safe_creation_tx_builder_with_not_enough_funds(self):
        w3 = self.w3
        s = generate_valid_s()
        funder_account = self.ethereum_test_account
        owners = [get_eth_address_with_key()[0] for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreationTx(w3=w3,
                                          owners=owners,
                                          threshold=threshold,
                                          signature_s=s,
                                          master_copy=self.safe_contract_V0_0_1_address,
                                          gas_price=gas_price,
                                          funder=NULL_ADDRESS)

        logger.info("Send %d gwei to deployer %s",
                    w3.fromWei(safe_creation_tx.payment_ether - 1, 'gwei'),
                    safe_creation_tx.deployer_address)
        self.send_tx({
            'to': safe_creation_tx.deployer_address,
            'value': safe_creation_tx.payment_ether - 1
        }, funder_account)

        with self.assertRaisesMessage(ValueError, 'enough funds'):
            w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)

    def test_safe_creation_tx_builder_with_payment(self):
        logger.info("Test Safe Proxy creation With Payment".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3

        s = generate_valid_s()

        funder_account = self.ethereum_test_account
        owners = [get_eth_address_with_key()[0] for _ in range(2)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreationTx(w3=w3,
                                          owners=owners,
                                          threshold=threshold,
                                          signature_s=s,
                                          master_copy=self.safe_contract_V0_0_1_address,
                                          gas_price=gas_price,
                                          funder=funder_account.address)

        user_external_account = Account.create()
        # Send some ether to that account
        safe_balance = w3.toWei(0.01, 'ether')
        self.send_tx({
            'to': user_external_account.address,
            'value': safe_balance * 2
        }, funder_account)

        logger.info("Send %d ether to safe %s", w3.fromWei(safe_balance, 'ether'), safe_creation_tx.safe_address)
        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_balance
        }, user_external_account)
        self.assertEqual(w3.eth.get_balance(safe_creation_tx.safe_address), safe_balance)

        logger.info("Send %d gwei to deployer %s", w3.fromWei(safe_creation_tx.payment_ether, 'gwei'),
                    safe_creation_tx.deployer_address)
        self.send_tx({
            'to': safe_creation_tx.deployer_address,
            'value': safe_creation_tx.payment_ether
        }, funder_account)

        logger.info("Create proxy contract with address %s", safe_creation_tx.safe_address)

        funder_balance = w3.eth.get_balance(funder_account.address)

        # This tx will create the Safe Proxy and return ether to the funder
        tx_hash = w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        self.assertEqual(tx_receipt.contractAddress, safe_creation_tx.safe_address)

        self.assertEqual(w3.eth.get_balance(funder_account.address),
                         funder_balance + safe_creation_tx.payment)

        logger.info("Deployer account has still %d gwei left (will be lost)",
                    w3.fromWei(w3.eth.get_balance(safe_creation_tx.deployer_address), 'gwei'))

        deployed_safe_proxy_contract = get_safe_contract(w3, tx_receipt.contractAddress)

        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)

    def test_safe_creation_tx_builder_with_token_payment(self):
        logger.info("Test Safe Proxy creation With Gas Payment".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3

        s = generate_valid_s()

        erc20_deployer = Account.create()
        funder_account = self.ethereum_test_account

        # Send something to the erc20 deployer
        self.send_tx({
            'to': erc20_deployer.address,
            'value': w3.toWei(1, 'ether')
        }, funder_account)

        funder = funder_account.address
        owners = [get_eth_address_with_key()[0] for _ in range(2)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        token_amount = int(1e18)
        erc20_contract = self.deploy_example_erc20(token_amount, erc20_deployer.address)
        self.assertEqual(erc20_contract.functions.balanceOf(erc20_deployer.address).call(), token_amount)

        safe_creation_tx = SafeCreationTx(w3=w3,
                                          owners=owners,
                                          threshold=threshold,
                                          signature_s=s,
                                          master_copy=self.safe_contract_V0_0_1_address,
                                          gas_price=gas_price,
                                          payment_token=erc20_contract.address,
                                          funder=funder)

        # In this test we will pretend that ether value = token value, so we send tokens as ether payment
        payment = safe_creation_tx.payment
        deployer_address = safe_creation_tx.deployer_address
        safe_address = safe_creation_tx.safe_address
        logger.info("Send %d tokens to safe %s", payment, safe_address)
        self.send_tx(erc20_contract.functions.transfer(safe_address,
                                                       payment).buildTransaction({'from': erc20_deployer.address}),
                     erc20_deployer)
        self.assertEqual(erc20_contract.functions.balanceOf(safe_address).call(), payment)

        logger.info("Send %d ether to deployer %s", w3.fromWei(payment, 'ether'), deployer_address)
        self.send_tx({
            'to': safe_creation_tx.deployer_address,
            'value': safe_creation_tx.payment
        }, funder_account)

        logger.info("Create proxy contract with address %s", safe_creation_tx.safe_address)
        funder_balance = w3.eth.get_balance(funder)

        # This tx will create the Safe Proxy and return tokens to the funder
        tx_hash = w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        self.assertEqual(tx_receipt.contractAddress, safe_address)
        self.assertEqual(w3.eth.get_balance(funder), funder_balance)
        self.assertEqual(erc20_contract.functions.balanceOf(funder).call(), payment)
        self.assertEqual(erc20_contract.functions.balanceOf(safe_address).call(), 0)

        logger.info("Deployer account has still %d gwei left (will be lost)",
                    w3.fromWei(w3.eth.get_balance(safe_creation_tx.deployer_address), 'gwei'))

        deployed_safe_proxy_contract = get_safe_contract(w3, tx_receipt.contractAddress)

        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)

        # Payment should be <= when payment_token_eth_value is 1.0
        # Funder will already have tokens so no storage need to be paid)
        safe_creation_tx_2 = SafeCreationTx(w3=w3,
                                            owners=owners,
                                            threshold=threshold,
                                            signature_s=s,
                                            master_copy=self.safe_contract_V0_0_1_address,
                                            gas_price=gas_price,
                                            payment_token=erc20_contract.address,
                                            payment_token_eth_value=1.0,
                                            funder=funder)
        self.assertLessEqual(safe_creation_tx_2.payment, safe_creation_tx.payment)

        # Now payment should be equal when payment_token_eth_value is 1.0
        safe_creation_tx_3 = SafeCreationTx(w3=w3,
                                            owners=owners,
                                            threshold=threshold,
                                            signature_s=s,
                                            master_copy=self.safe_contract_V0_0_1_address,
                                            gas_price=gas_price,
                                            payment_token=erc20_contract.address,
                                            payment_token_eth_value=1.0,
                                            funder=funder)
        self.assertEqual(safe_creation_tx_3.payment, safe_creation_tx_2.payment)

        # Check that payment is less when payment_token_eth_value is set(token value > ether)
        safe_creation_tx_4 = SafeCreationTx(w3=w3,
                                            owners=owners,
                                            threshold=threshold,
                                            signature_s=s,
                                            master_copy=self.safe_contract_V0_0_1_address,
                                            gas_price=gas_price,
                                            payment_token=erc20_contract.address,
                                            payment_token_eth_value=1.1,
                                            funder=funder)
        self.assertLess(safe_creation_tx_4.payment, safe_creation_tx.payment)

        # Check that payment is more when payment_token_eth_value is set(token value < ether)
        safe_creation_tx_5 = SafeCreationTx(w3=w3,
                                            owners=owners,
                                            threshold=threshold,
                                            signature_s=s,
                                            master_copy=self.safe_contract_V0_0_1_address,
                                            gas_price=gas_price,
                                            payment_token=erc20_contract.address,
                                            payment_token_eth_value=0.1,
                                            funder=funder)
        self.assertGreater(safe_creation_tx_5.payment, safe_creation_tx.payment)

    def test_safe_creation_tx_builder_with_fixed_cost(self):
        logger.info("Test Safe Proxy creation With Fixed Cost".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3

        s = generate_valid_s()

        funder_account = self.ethereum_test_account
        owners = [get_eth_address_with_key()[0] for _ in range(2)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        fixed_creation_cost = 123  # Wei

        safe_creation_tx = SafeCreationTx(w3=w3,
                                          owners=owners,
                                          threshold=threshold,
                                          signature_s=s,
                                          master_copy=self.safe_contract_V0_0_1_address,
                                          gas_price=gas_price,
                                          payment_token=None,
                                          funder=funder_account.address,
                                          fixed_creation_cost=fixed_creation_cost)

        self.assertEqual(safe_creation_tx.payment, fixed_creation_cost)
        self.assertEqual(safe_creation_tx.payment_ether, safe_creation_tx.gas * safe_creation_tx.gas_price)

        deployer_address = safe_creation_tx.deployer_address
        safe_address = safe_creation_tx.safe_address
        safe_balance = w3.toWei(0.01, 'ether')
        logger.info("Send %d ether to safe %s", w3.fromWei(safe_balance, 'ether'), safe_address)
        self.send_tx({
            'to': safe_address,
            'value': safe_balance
        }, funder_account)
        self.assertEqual(w3.eth.get_balance(safe_address), safe_balance)

        logger.info("Send %d ether to deployer %s", w3.fromWei(safe_creation_tx.payment_ether, 'ether'),
                    deployer_address)
        self.send_tx({
            'to': deployer_address,
            'value': safe_creation_tx.payment_ether
        }, funder_account)

        logger.info("Create proxy contract with address %s", safe_creation_tx.safe_address)

        funder_balance = w3.eth.get_balance(funder_account.address)

        # This tx will create the Safe Proxy and return tokens to the funder
        tx_hash = w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        self.assertEqual(tx_receipt.contractAddress, safe_address)
        self.assertEqual(w3.eth.get_balance(safe_address), safe_balance - fixed_creation_cost)
        self.assertLess(w3.eth.get_balance(deployer_address), safe_creation_tx.payment_ether)
        self.assertEqual(w3.eth.get_balance(funder_account.address), funder_balance + safe_creation_tx.payment)

        logger.info("Deployer account has still %d gwei left (will be lost)",
                    w3.fromWei(w3.eth.get_balance(safe_creation_tx.deployer_address), 'gwei'))

        deployed_safe_proxy_contract = get_safe_contract(w3, safe_address)

        self.assertEqual(deployed_safe_proxy_contract.functions.getThreshold().call(), threshold)
        self.assertEqual(deployed_safe_proxy_contract.functions.getOwners().call(), owners)

    def test_safe_gas_with_multiple_owners(self):
        logger.info("Test Safe Proxy creation gas with multiple owners".center(LOG_TITLE_WIDTH, '-'))
        w3 = self.w3
        funder_account = self.ethereum_test_account
        number_of_accounts = 10
        for i in range(2, number_of_accounts):
            s = generate_valid_s()
            owners = [get_eth_address_with_key()[0] for _ in range(i + 1)]
            threshold = len(owners)
            gas_price = w3.toWei(15, 'gwei')

            safe_creation_tx = SafeCreationTx(w3=w3,
                                              owners=owners,
                                              threshold=threshold,
                                              signature_s=s,
                                              master_copy=self.safe_contract_V0_0_1_address,
                                              gas_price=gas_price,
                                              funder=None)

            self.send_tx({
                'to': safe_creation_tx.deployer_address,
                'value': safe_creation_tx.payment
            }, funder_account)
            tx_hash = w3.eth.send_raw_transaction(safe_creation_tx.tx_raw)
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
            self.assertEqual(tx_receipt.contractAddress, safe_creation_tx.safe_address)

            logger.info("Number of owners: %d - Gas estimated %d - Gas Used %d - Difference %d",
                        len(owners),
                        safe_creation_tx.gas,
                        tx_receipt.gasUsed,
                        safe_creation_tx.gas - tx_receipt.gasUsed)



from typing import List

from django.test import TestCase

from ...clients import Sourcify
from ...ethereum_client import EthereumNetwork


class TestSourcify(TestCase):
    def test_sourcify_get_contract_metadata(self):
        sourcify = Sourcify()
        safe_contract_address = '0x6851D6fDFAfD08c0295C392436245E5bc78B0185'
        try:
            contract_metadata = sourcify.get_contract_metadata(safe_contract_address)
        except IOError:
            self.skipTest('Cannot connect to Sourcify')
        self.assertEqual(contract_metadata.name, 'GnosisSafe')
        self.assertIsInstance(contract_metadata.abi, List)
        self.assertTrue(contract_metadata.abi)
        self.assertFalse(contract_metadata.partial_match)
        contract_metadata_rinkeby = Sourcify(EthereumNetwork.RINKEBY).get_contract_metadata(safe_contract_address)
        self.assertEqual(contract_metadata, contract_metadata_rinkeby)

        partial_match_contract_address = '0x000000000000C1CB11D5c062901F32D06248CE48'
        contract_metadata = sourcify.get_contract_metadata(partial_match_contract_address)
        self.assertEqual(contract_metadata.name, 'LiquidGasToken')
        self.assertIsInstance(contract_metadata.abi, List)
        self.assertTrue(contract_metadata.abi)
        self.assertTrue(contract_metadata.partial_match)



import json
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests
from eth_typing import ChecksumAddress

from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class BlockscoutClientException(Exception):
    pass


class BlockScoutConfigurationProblem(BlockscoutClientException):
    pass


class BlockscoutClient:
    NETWORK_WITH_URL = {
        EthereumNetwork.XDAI: 'https://blockscout.com/poa/xdai/',
        EthereumNetwork.MATIC: 'https://polygon-explorer-mainnet.chainstacklabs.com/',
        EthereumNetwork.MUMBAI: 'https://polygon-explorer-mumbai.chainstacklabs.com/',
        EthereumNetwork.ENERGY_WEB_CHAIN: 'https://explorer.energyweb.org/',
        EthereumNetwork.VOLTA: 'https://volta-explorer.energyweb.org/',
    }

    def __init__(self, network: EthereumNetwork):
        self.network = network
        self.base_url = self.NETWORK_WITH_URL.get(network)
        if self.base_url is None:
            raise BlockScoutConfigurationProblem(f'Network {network.name} - {network.value} not supported')
        self.grahpql_url = self.base_url + '/graphiql'
        self.http_session = requests.Session()

    def build_url(self, path: str):
        return urljoin(self.base_url, path)

    def _do_request(self, url: str, query: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.post(url, json={'query': query}, timeout=10)
        if not response.ok:
            return None

        return response.json()

    def get_contract_metadata(self, address: ChecksumAddress) -> Optional[ContractMetadata]:
        query = '{address(hash: "%s") { hash, smartContract {name, abi} }}' % address
        result = self._do_request(self.grahpql_url, query)
        if result and 'error' not in result and result.get('data', {}).get('address', {}) \
                and result['data']['address']['smartContract']:
            smart_contract = result['data']['address']['smartContract']
            return ContractMetadata(smart_contract['name'], json.loads(smart_contract['abi']), False)



import json
import os
import sys
from typing import Any, Dict, Optional

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract


def load_contract_interface(file_name):
    return _load_json_file(_abi_file_path(file_name))


def _abi_file_path(file):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), file))


def _load_json_file(path):
    with open(path) as f:
        return json.load(f)


current_module = sys.modules[__name__]

"""
Safe Addresses. Should be the same for every chain. Check:
https://github.com/gnosis/safe-contracts/blob/development/zos.mainnet.json
https://github.com/gnosis/safe-contracts/blob/development/zos.rinkeby.json

GnosisSafeV1.1.1: 0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F
GnosisSafeV1.1.0: 0xaE32496491b53841efb51829d6f886387708F99B
GnosisSafeV1.0.0: 0xb6029EA3B2c51D09a50B53CA8012FeEB05bDa35A

Factories
ProxyFactoryV1.1.0: 0x50e55Af101C777bA7A1d560a774A82eF002ced9F
ProxyFactoryV1.0.0: 0x12302fE9c02ff50939BaAaaf415fc226C078613C

Libraries
CreateAndAddModules: 0x1a56aE690ab0818aF5cA349b7D21f1d7e76a3d36
MultiSend: 0xD4B7B161E4779629C2717385114Bf78D612aEa72
"""

contracts = {
    'safe': 'GnosisSafe_V1_1_1.json',
    'safe_V1_3_0': 'GnosisSafe_V1_3_0.json',
    'safe_V1_0_0': 'GnosisSafe_V1_0_0.json',
    'safe_V0_0_1': 'GnosisSafe_V0_0_1.json',
    'erc20': 'ERC20.json',
    'erc721': 'ERC721.json',
    'example_erc20': 'ERC20TestToken.json',
    'delegate_constructor_proxy': 'DelegateConstructorProxy.json',
    'multi_send': 'MultiSend.json',
    'paying_proxy': 'PayingProxy.json',
    'proxy_factory': 'ProxyFactory_V1_3_0.json',
    'proxy_factory_V1_1_1': 'ProxyFactory_V1_1_1.json',
    'proxy_factory_V1_0_0': 'ProxyFactory_V1_0_0.json',
    'proxy': 'Proxy_V1_1_1.json',
    'uniswap_exchange': 'uniswap_exchange.json',
    'uniswap_factory': 'uniswap_factory.json',
    'uniswap_v2_factory': 'uniswap_v2_factory.json',
    'uniswap_v2_pair': 'uniswap_v2_pair.json',
    'uniswap_v2_router': 'uniswap_v2_router.json',  # Router02
    'kyber_network_proxy': 'kyber_network_proxy.json',
    'cpk_factory': 'CPKFactory.json',
}


def generate_contract_fn(contract: Dict[str, Any]):
    """
    Dynamically generate functions to work with the contracts
    :param contract:
    :return:
    """
    def fn(w3: Web3, address: Optional[ChecksumAddress] = None):
        return w3.eth.contract(address=address,
                               abi=contract['abi'],
                               bytecode=contract.get('bytecode'))
    return fn


# Anotate functions that will be generated later with `setattr` so typing does not complains
def get_safe_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_safe_V1_3_0_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_safe_V1_0_0_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_safe_V0_0_1_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_erc20_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_erc721_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_example_erc20_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_delegate_constructor_proxy_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_multi_send_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_paying_proxy_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_proxy_factory_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_proxy_factory_V1_1_1_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_proxy_factory_V1_0_0_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_proxy_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_uniswap_exchange_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_uniswap_factory_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_uniswap_v2_factory_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_uniswap_v2_pair_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_uniswap_v2_router_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_kyber_network_proxy_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_cpk_factory_contract(w3: Web3, address: Optional[str] = None) -> Contract:
    pass


def get_paying_proxy_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface('PayingProxy.json')['deployedBytecode'])


def get_proxy_1_0_0_deployed_bytecode() -> bytes:
    return HexBytes(load_contract_interface('Proxy_V1_0_0.json')['deployedBytecode'])


for contract_name, json_contract_filename in contracts.items():
    fn_name = 'get_{}_contract'.format(contract_name)
    contract_dict = load_contract_interface(json_contract_filename)
    setattr(current_module, fn_name, generate_contract_fn(contract_dict))



import os
from typing import Tuple, Union

import eth_abi
from ethereum import utils
from hexbytes import HexBytes
from web3 import Web3


def get_eth_address_with_key() -> Tuple[str, bytes]:
    # import secp256k1
    # private_key = secp256k1.PrivateKey().private_key
    private_key = utils.sha3(os.urandom(4096))
    public_key = utils.checksum_encode(utils.privtoaddr(private_key))
    # If you want to use secp256k1 to calculate public_key
    # utils.checksum_encode(utils.sha3(p.pubkey.serialize(compressed=False)[1:])[-20:])

    return public_key, private_key


def get_eth_address_with_invalid_checksum() -> str:
    address, _ = get_eth_address_with_key()
    return '0x' + ''.join([c.lower() if c.isupper() else c.upper() for c in address[2:]])


def generate_address_2(from_: Union[str, bytes], salt: Union[str, bytes], init_code: Union[str, bytes]) -> str:
    """
    Generates an address for a contract created using CREATE2.

    :param from_: The address which is creating this new address (need to be 20 bytes)
    :param salt: A salt (32 bytes)
    :param init_code: A init code of the contract being created
    :return: Address of the new contract
    """

    from_ = HexBytes(from_)
    salt = HexBytes(salt)
    init_code = HexBytes(init_code)

    assert len(from_) == 20, f'Address {from_.hex()} is not valid. Must be 20 bytes'
    assert len(salt) == 32, f'Salt {salt.hex()} is not valid. Must be 32 bytes'
    assert len(init_code) > 0, f'Init code {init_code.hex()} is not valid'

    init_code_hash = Web3.keccak(init_code)
    contract_address = Web3.keccak(HexBytes('ff') + from_ + salt + init_code_hash)
    return Web3.toChecksumAddress(contract_address[12:])


def decode_string_or_bytes32(data: bytes) -> str:
    try:
        return eth_abi.decode_abi(['string'], data)[0]
    except OverflowError:
        name = eth_abi.decode_abi(['bytes32'], data)[0]
        end_position = name.find(b'\x00')
        if end_position == -1:
            return name.decode()
        else:
            return name[:end_position].decode()


def remove_swarm_metadata(code: bytes) -> bytes:
    """
    Remove swarm metadata from Solidity bytecode
    :param code:
    :return: Code without metadata
    """
    swarm = b'\xa1\x65bzzr0'
    position = code.find(swarm)
    if position == -1:
        raise ValueError('Swarm metadata not found in code %s' % code.hex())
    return code[:position]


def compare_byte_code(code_1: bytes, code_2: bytes) -> bool:
    """
    Compare code, removing swarm metadata if necessary
    :param code_1:
    :param code_2:
    :return: True if same code, False otherwise
    """
    if code_1 == code_2:
        return True
    else:
        codes = []
        for code in (code_1, code_2):
            try:
                codes.append(remove_swarm_metadata(code))
            except ValueError:
                codes.append(code)

        return codes[0] == codes[1]



from django.test import TestCase

from gnosis.eth.constants import (SIGNATURE_R_MAX_VALUE, SIGNATURE_R_MIN_VALUE,
                                  SIGNATURE_S_MAX_VALUE, SIGNATURE_S_MIN_VALUE)
from gnosis.eth.utils import get_eth_address_with_key

from ..serializers import (SafeMultisigEstimateTxSerializer,
                           SafeSignatureSerializer)


class TestSerializers(TestCase):
    def test_safe_signature_serializer(self):
        for v in [0, 1]:
            self.assertFalse(SafeSignatureSerializer(data={'v': v, 'r': -1, 's': 0}).is_valid())
            self.assertTrue(SafeSignatureSerializer(data={'v': v, 'r': 0, 's': 0}).is_valid())
            self.assertTrue(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MAX_VALUE + 1, 's': 0}).is_valid())
            self.assertTrue(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MAX_VALUE + 1,
                                                          's': SIGNATURE_S_MAX_VALUE + 1}).is_valid())

        for v in [27, 28]:
            self.assertFalse(SafeSignatureSerializer(data={'v': v, 'r': 0, 's': 0}).is_valid())
            self.assertTrue(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MIN_VALUE + 1,
                                                          's': SIGNATURE_S_MAX_VALUE - 1}).is_valid())
            self.assertTrue(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MAX_VALUE - 1,
                                                          's': SIGNATURE_S_MIN_VALUE + 1}).is_valid())

            self.assertFalse(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MAX_VALUE + 1,
                                                           's': SIGNATURE_S_MAX_VALUE - 1}).is_valid())
            self.assertFalse(SafeSignatureSerializer(data={'v': v, 'r': SIGNATURE_R_MIN_VALUE + 1,
                                                           's': SIGNATURE_S_MAX_VALUE + 1}).is_valid())

    def test_safe_multisig_tx_estimate_serializer(self):
        safe_address, _ = get_eth_address_with_key()
        eth_address, _ = get_eth_address_with_key()
        data = {
            'safe': safe_address,
            'to': None,
            'data': None,
            'value': 1,
            'operation': 0
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)

        # To and data cannot be empty
        self.assertFalse(serializer.is_valid())

        data = {
            'safe': safe_address,
            'to': eth_address,
            'data': '0x00',
            'value': 1,
            'operation': 2
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)
        # Operation cannot be contract creation and to set
        self.assertFalse(serializer.is_valid())

        # Create Operation has been disabled
        data = {
            'safe': safe_address,
            'to': None,
            'data': '0x00',
            'value': 1,
            'operation': 2
        }
        serializer = SafeMultisigEstimateTxSerializer(data=data)
        # Operation is not contract creation and to is not empty
        self.assertFalse(serializer.is_valid())



from django.test import TestCase

from ... import EthereumNetwork
from ...clients import EtherscanClient, EtherscanRateLimitError
from .mocks import sourcify_safe_metadata


class TestEtherscanClient(TestCase):
    def test_etherscan_get_abi(self):
        try:
            etherscan_api = EtherscanClient(EthereumNetwork.MAINNET)
            safe_master_copy_abi = sourcify_safe_metadata['output']['abi']
            safe_master_copy_address = '0x6851D6fDFAfD08c0295C392436245E5bc78B0185'
            self.assertEqual(etherscan_api.get_contract_abi(safe_master_copy_address),
                             safe_master_copy_abi)

            contract_metadata = etherscan_api.get_contract_metadata(safe_master_copy_address)
            self.assertEqual(contract_metadata.name, 'GnosisSafe')
            self.assertEqual(contract_metadata.abi, safe_master_copy_abi)

            random_address = '0xaE32496491b53841efb51829d6f886387708F99a'
            self.assertIsNone(etherscan_api.get_contract_abi(random_address))
            self.assertIsNone(etherscan_api.get_contract_metadata(random_address))
        except EtherscanRateLimitError:
            self.skipTest('Etherscan rate limit reached')



# flake8: noqa F401
from .oracles import (AaveOracle, BalancerOracle, CannotGetPriceFromOracle,
                      ComposedPriceOracle, CurveOracle, InvalidPriceFromOracle,
                      KyberOracle, MooniswapOracle, OracleException,
                      PoolTogetherOracle, PriceOracle, PricePoolOracle,
                      SushiswapOracle, UnderlyingToken, UniswapOracle,
                      UniswapV2Oracle, UsdPricePoolOracle, YearnOracle)
